/*
 * SPDX-FileCopyrightText: 2021 Andrea Pappacoda
 *
 * SPDX-License-Identifier: AGPL-3.0-or-later
 */

#pragma once
#define BOOST_UT_DISABLE_MODULE
#include <boost/ut.hpp>
#include <ddns/cloudflare-ddns.h>
#include "credentials.hpp"

using namespace boost::ut;
